"""Script to generate doxygen documentation.
"""

import re
import os
import os.path
import sys

def find_program(filename):
    """find a program in folders path_lst, and sets env[var]
    @param env: environmentA
    @param filename: name of the program to search for
    @param path_list: list of directories to search for filename
    @param var: environment value to be checked for in env or os.environ
    @return: either the value that is referenced with [var] in env or os.environ
    or the first occurrence filename or '' if filename could not be found
"""
    paths = os.environ.get('PATH', '').split(os.pathsep)
    suffixes = ('win32' in sys.platform ) and '.exe .com .bat .cmd' or ''
    for name in [filename+ext for ext in suffixes.split()]:
        for directory in paths:
            full_path = os.path.join(directory, name)
            if os.path.isfile(full_path):
                return full_path
    return ''

def do_subst_in_file(targetfile, sourcefile, dict):
    """Replace all instances of the keys of dict with their values.
    For example, if dict is {'%VERSION%': '1.2345', '%BASE%': 'MyProg'},
    then all instances of %VERSION% in the file will be replaced with 1.2345 etc.
    """
    try:
        f = open(sourcefile, 'rb')
        contents = f.read()
        f.close()
    except:
        print "Can't read source file %s"%sourcefile
        raise
    for (k,v) in dict.items():
        v = v.replace('\\','\\\\') 
        contents = re.sub(k, v, contents)
    try:
        f = open(targetfile, 'wb')
        f.write(contents)
        f.close()
    except:
        print "Can't write target file %s"%targetfile
        raise

def run_doxygen(doxygen_path, config_file, working_dir):
    config_file = os.path.abspath( config_file )
    doxygen_path = doxygen_path
    os.chdir( working_dir )
    cmd = [doxygen_path, config_file]
    print ' '.join( cmd )
    try:
        import subprocess
    except:
        if os.system( ' '.join( cmd ) ) != 0:
            print 'Documentation generation failed'
            return False
    else:
        try:
            subprocess.check_call( cmd )
        except subprocess.CalledProcessError:
            return False
    return True

def main():
    usage = """%prog
    Generates doxygen documentation.
    """
    from optparse import OptionParser
    parser = OptionParser(usage=usage)
    parser.allow_interspersed_args = False
    parser.add_option('--with-dot', dest="with_dot", action='store_true', default=False,
        help="""Enable usage of DOT to generate collaboration diagram""")
    parser.add_option('--dot', dest="dot_path", action='store', default=find_program('dot'),
        help="""Path to GraphViz dot tool. Must be full qualified path. [Default: %default]""")
    parser.add_option('--doxygen', dest="doxygen_path", action='store', default=find_program('doxygen'),
        help="""Path to Doxygen tool. [Default: %default]""")
    parser.add_option('--with-html-help', dest="with_html_help", action='store_true', default=False,
        help="""Enable generation of Microsoft HTML HELP""")
    parser.add_option('--no-uml-look', dest="with_uml_look", action='store_false', default=True,
        help="""Generates DOT graph without UML look [Default: False]""")
    parser.add_option('--open', dest="open", action='store_true', default=False,
        help="""Open the HTML index in the web browser after generation""")
    parser.enable_interspersed_args()
    options, args = parser.parse_args()

    version = open('version','rt').read().strip()
    output_dir = '../build/doxygen' # relative to doc/doxyfile location.
    def yesno( bool ):
        return bool and 'YES' or 'NO'
    subst_keys = {
        '%CPPUNIT_VERSION%': version,
        '%DOC_TOPDIR%': '',
        '%TOPDIR%': os.path.abspath('.'),
        '%HTML_OUTPUT%': os.path.join( output_dir, 'cppunit2-doc-html-' + version ),
        '%HAVE_DOT%': yesno(options.with_dot),
        '%DOT_PATH%': os.path.split(options.dot_path)[0],
        '%HTML_HELP%': yesno(options.with_html_help),
        '%UML_LOOK%': yesno(options.with_uml_look)
        }

    full_output_dir = os.path.join( 'doc', output_dir )
    if not os.path.isdir( full_output_dir ):
        os.makedirs( full_output_dir )

    do_subst_in_file( 'doc/doxyfile', 'doc/doxyfile.in', subst_keys )
    ok = run_doxygen( options.doxygen_path, 'doc/doxyfile', 'doc' )
    print open('cppunit2-doxygen-warning.log','rb').read()
    if not ok:
        print 'Doxygen generation failed'
    index_path = os.path.abspath(os.path.join(subst_keys['%HTML_OUTPUT%'], 'index.html'))
    print 'Generated documentation can be found in:'
    print index_path
    if options.open:
        import webbrowser
        webbrowser.open( 'file://' + index_path )

if __name__ == '__main__':
    main()
