#ifndef LIBCPPUT_IPC_H_INCLUDED
# define LIBCPPUT_IPC_H_INCLUDED

# include <cpput/forwards.h>
# include <cpptl/conststring.h>
# include <json/value.h>
# include <memory>
# include <deque>


/* Inter-process Communication.
 * Synchronous write, Asynchonous read.
 * Synchronous write is required to ensure that all message that are send from the test runner
 * will not be lost (so that the test that was running when a crash occurred is known).
 */

namespace CppUT {


//class CPPUT_API ReadEventHandler
//{
//public:
//   virtual ~ReadEventHandler();
//
//   /** Called when data are read from the buffer.
//    */
//   virtual void onRead( void *buffer, size_t length ) = 0;
//
//   /** Called when the input stream has been closed.
//    */
//   virtual void onClose() = 0;
//};


class CPPUT_API SynchronousWriteChannel : private CppTL::NonCopyable
{
public:
   virtual ~SynchronousWriteChannel();

   virtual bool read( void *readBuffer, size_t bufferSize, size_t &readLength ) = 0;

   virtual bool write( const void *buffer, size_t length ) = 0;

   virtual void closeRead() = 0;

   virtual void closeWrite() = 0;

   /// Close read & write channel.
   void close();
};

typedef std::auto_ptr<SynchronousWriteChannel> SynchronousWriteChannelAutoPtr;

SynchronousWriteChannelAutoPtr makeSlaveChannel();

/*! \brief Represents the identifier of a JSON RPC request.
 */
class CPPUT_API JsonRPCRequest
{
public:
   explicit JsonRPCRequest( const Json::Value &requestId );

   const Json::Value &requestId() const;

private:
   Json::Value id_;
};


/*! \brief Handles JSON RPC request and notification.
 */
class CPPUT_API JsonRPCRequestHandler : private CppTL::NonCopyable
{
public:
   virtual ~JsonRPCRequestHandler();

   /*! \brief Called when a request is received.
    * @param request Identifier of the request required to pass to JsonRPCChannel::reply or JsonRPCChannel::replyWithError.
    * @param method Name of the invoked method.
    * @param parameter A JSON Object that contains name/value parameters.
    * @param channel Channel to use to reply.
    */
   virtual void onRequest( JsonRPCRequest request,
                           const std::string &method,
                           const Json::Value &parameter,
                           JsonRPCChannel &channel ) = 0;

   /*! \brief Called when a notification is received.
    * @param method Name of the notification method.
    * @param parameter A JSON Object that contains name/value parameters.
    * @param channel Channel to use to send another message.
    * Notes: you can not reply to notification.
    */
   virtual void onNotification( const std::string &method, 
                                const Json::Value &parameter, 
                                JsonRPCChannel &channel ) = 0;
};


/*! \brief Represents a JSON RPC error.
 */
class CPPUT_API JsonRPCError
{
public:
   static JsonRPCError parseError;
   static JsonRPCError invalidRequest;
   static JsonRPCError internalError;

   static JsonRPCError methodNotFound;
   static JsonRPCError invalidParameter;

   JsonRPCError( int errorCode, const char *message );
   JsonRPCError( int errorCode, const char *message, const Json::Value &detail );

   JsonRPCError detail( const Json::Value &detail );

   int errorCode() const;
   const char *message() const;
   const Json::Value &detail() const;

private:
   Json::Value detail_;
   CppTL::ConstString message_;
   int errorCode_;
};

/*! \brief Channel to receives and send message using JSON RPC.
 */
class CPPUT_API JsonRPCChannel : private CppTL::NonCopyable
{
public:
   /*! \brief Constructs the channel using the specified transport and request handler.
    */
   JsonRPCChannel( SynchronousWriteChannelAutoPtr channel,
                   JsonRPCRequestHandler &requestHandler );
   
   virtual ~JsonRPCChannel();

   /*! \brief Close the underlying transport.
    */
   void close();

   /*! \brief Wait for requests and handle them.
    * Loop until either the underlying transport is closed, or stopWaitingForRequests() is called.
    */
   void waitForRequests();

   /*! \brief Stops waiting for request.
    * Cause function waitForRequests() to exit as soon as the request handler returns.
    */
   void stopWaitingForRequests();

   /*! \brief Sends a request to the peer.
    * @param method Name of the invoked method.
    * @param parameter JSON Object containing name/value parameters.
    * Request have a response with either a result or an error.
    */
   void sendRequest( const char *method, const Json::Value &parameter );

   /*! \brief Sends a notification to the peer.
    * @param method Name of the invoked method.
    * @param parameter JSON Object containing name/value parameters.
    * Notification do not have a response.
    */
   void sendNotification( const char *method, const Json::Value &parameter );

   /*! \brief Reply to a request.
    * @param request Identifier of the request being replied to.
    * @param result JSON object representing the result of the request.
    * Usually called from JsonRPCRequestHandler::onRequest() but may be called at any time as long as a valid request
    * is provided.
    */
   void reply( JsonRPCRequest request,
               const Json::Value &result );

   /*! \brief Reply to a request indicating an error.
    * @param request Identifier of the request being replied to.
    * @param error Error to provide to reply the request.
    */
   void replyWithError( JsonRPCRequest request,
                        const JsonRPCError &error );

private:
   /*! Sends a JSON RPC object to the peer.
    */
   void sendMessage( const Json::Value &message );

   void decodeIncommingData( const char *begin, size_t length );
   void processIncommingData();

   SynchronousWriteChannelAutoPtr channel_;
   JsonRPCRequestHandler &requestHandler_;
   Json::Value::Int nextRequestId_;
   typedef std::deque<CppTL::ConstString> PendingRequests;
   PendingRequests pendingRequests_;
   std::string incommingData_;
   bool stopWaiting_;
};


} // namespace CppUT {

#endif // LIBCPPUT_IPC_H_INCLUDED
