#ifndef CPPUTTOOLS_CONFIGURATION_DESCRIPTION_H_INCUDED
# define CPPUTTOOLS_CONFIGURATION_DESCRIPTION_H_INCUDED

# include "forwards.h"
# include "errors.h"
# include "validator.h"
# include <cpputtools/valuebinder.h>

# include <iostream>
# include <map>
# include <set>

namespace CppUTTools { namespace CommandLineOptions {


class OptionDescription // not copyable
{
public:
   enum ParameterRequirement
   {
      noParameter,         // |
      optionalParameter,   // ?
      requiredParameter,   // :
      oneOrMoreParameter,  // +
      zeroOrMoreParameter  // *
   };

   OptionDescription( const Char &shortName, 
                      const String &longName,
                      const String &description,
                      const ValidatorPtr &validator );

   // @param packedName Has the following format: "longName,shortName". 
   //                   ShortName must be a single letter. Example: "exclude,x"
   //                   Will match --exclude or -x on the command line.
   OptionDescription( const String &packedName,
                      const String &description,
                      const ValidatorPtr &validator );

   ~OptionDescription();

   OptionDescription &setValidator( const ValidatorPtr &validator );

   // Switch parameter requirement as follow:
   // * if value is empty:
   // optionalParameter -> requiredParameter
   // * if value is not empty:
   // noParameter, requiredParameter -> optionalParameter
   // oneOrMoreParameter -> zeroOrMoreParameter
   OptionDescription &setDefaultValue( const Value &value );

   OptionDescription &setStorage( const ValueBinder &storage );

   const Value &defaultValue() const;

   bool hasDefaultValue() const;

   bool hasStorage() const;

   bool hasShortName() const;

   bool hasLongName() const;

   String shortName() const;

   String longName() const;

   String name() const;

   String description() const;

   bool hasNoParameter() const;

   bool mayHaveParameter() const;

   bool requireParameter() const;

   // @exception ValidatorError if validation fail.
   Value validate( const Parameters &parameters ) const;

   // Store the final value into the location specified by the user if any.
   void store( const Value &value ) const;

private:
   void setDefaultParameterRequirement();

   Value defaultValue_;
   String shortName_;
   String longName_;
   ValueBinder storage_;
   String description_;
   ParameterRequirement parameter_;
   ValidatorPtr validator_;
};


class OptionsDescription
{
public:
   OptionsDescription( const String &caption );
   virtual ~OptionsDescription();

   /// @exception OptionsDescriptionError if an option has already been registered with conflicting
   ///                                    names.
   OptionsDescription &add( const OptionDescription &description );

   /// @exception OptionsDescriptionError if an option has already been registered with conflicting
   ///                                    names.
   OptionsDescription &add( const String &optionName,
                            const String &description );


#if 0
   /// @exception OptionsDescriptionError if an option has already been registered with conflicting
   ///                                    names.
   OptionsDescription &add( const String &optionName,
                            const OptionHandlerHelper &handlerHelper,
                            const String &description );

   /// @exception OptionsDescriptionError if an option has already been registered with conflicting
   ///                                    names.
   OptionsDescription &add( const String &optionName,
                            const OptionHandlerHelper &handlerHelper,
                            Validator *validator,
                            const String &description );

   /// @exception OptionsDescriptionError if an option has already been registered with conflicting
   ///                                    names.
   OptionsDescription &add( const String &optionName,
                            Validator *validator,
                            const String &description );

   /// @exception OptionsDescriptionError if an option has already been registered with conflicting
   ///                                    names.
   OptionsDescription &add( const String &optionName,
                            OptionHandler *handler,
                            const String &description );

   /// @exception OptionsDescriptionError if an option has already been registered with conflicting
   ///                                    names.
   OptionsDescription &add( const String &optionName,
                            OptionHandler *handler,
                            Validator *validator,
                            const String &description );
#endif

   std::set<String> optionNames() const;

   bool hasOption( const String &optionName ) const;

   // long name or -shortName
   const OptionDescription &description( const String &optionName ) const;

   friend std::ostream &operator <<( std::ostream &os, 
                                     const OptionsDescription &options );

private:
   typedef std::deque<OptionDescription> Options;
   Options options_;
   String caption_;

   typedef std::map<String,int> OptionIndexes;
   OptionIndexes shortNameIndexes_;
   OptionIndexes longNameIndexes_;
   OptionIndexes nameIndexes_;
};


template<class ValueType>
OptionDescription option( const std::string &optionName,
                          const std::string &description );

template<class DefaultValueType>
OptionDescription optionDefault( const std::string &optionName,
                                 const DefaultValueType &defaultValue,
                                 const std::string &description );

template<class StorageValueType>
OptionDescription option( const std::string &optionName,
                          StorageValueType &storageValue,
                          const std::string &description );

template<class StorageValueType
        ,class DefaultValueType>
OptionDescription option( const std::string &optionName,
                          StorageValueType &storageValue,
                          const DefaultValueType &defaultValue,
                          const std::string &description );


template<class ValueType>
OptionDescription validatedOption( const std::string &optionName,
                                   const std::string &description,
                                   const ValidatorPtr &validator );

template<class DefaultValueType>
OptionDescription validatedOptionDefault( const std::string &optionName,
                                          const DefaultValueType &defaultValue,
                                          const std::string &description,
                                          const ValidatorPtr &validator );

template<class StorageValueType>
OptionDescription validatedOption( const std::string &optionName,
                                   StorageValueType &storageValue,
                                   const std::string &description,
                                   const ValidatorPtr &validator );

template<class StorageValueType
        ,class DefaultValueType>
OptionDescription validatedOption( const std::string &optionName,
                                   StorageValueType &storageValue,
                                   const DefaultValueType &defaultValue,
                                   const std::string &description,
                                   const ValidatorPtr &validator );



// class OptionDescription
// //////////////////////////////////////////////////////////////////

inline
OptionDescription::OptionDescription( const String &packedName,
                                      const String &description,
                                      const ValidatorPtr &validator )
   : description_( description )
   , validator_( validator )
{
   unsigned int commaPos = packedName.find( ',' );
   String parameter( "|?:+*" );
   String cleanPackedName = packedName;
   Char parameterRequirement = cleanPackedName.at( cleanPackedName.length()-1 );
   int pos = parameter.find( parameterRequirement );
   if ( pos != String::npos )
   {
      static ParameterRequirement parameters[] = { 
         noParameter, optionalParameter, requiredParameter, oneOrMoreParameter, zeroOrMoreParameter
         };
      parameter_ = parameters[pos];
      cleanPackedName.resize( cleanPackedName.length()-1 );
   }
   else
      setDefaultParameterRequirement();

   if ( commaPos != String::npos )
   {
      longName_ = cleanPackedName.substr(0,commaPos);
      shortName_ = cleanPackedName.substr(commaPos+1,1);
   }
   else
      longName_ = cleanPackedName;

}


inline 
OptionDescription::OptionDescription( const Char &shortName, 
                                      const String &longName,
                                      const String &description,
                                      const ValidatorPtr &validator )
   : shortName_( String(1,shortName) )
   , longName_( longName )
   , description_( description )
   , validator_( validator )
{
   setDefaultParameterRequirement();
}


inline void 
OptionDescription::setDefaultParameterRequirement()
{
   if ( hasStorage()  ||  hasDefaultValue() )
      parameter_ = hasDefaultValue() ? optionalParameter 
                                     : requiredParameter;
   else
      parameter_ = noParameter;
}


inline 
OptionDescription::~OptionDescription()
{
}


inline OptionDescription &
OptionDescription::setValidator( const ValidatorPtr &validator )
{
   validator_ = validator;
   return *this;
}


inline OptionDescription &
OptionDescription::setDefaultValue( const Value &value )
{
   defaultValue_ = value;
   if ( value.isEmpty() )
   {
      if ( parameter_ == optionalParameter )
         parameter_ = requiredParameter;
   }
   else
   {
      if ( parameter_ == noParameter  ||  parameter_ == requiredParameter )
         parameter_ = optionalParameter;
      else if ( parameter_ == oneOrMoreParameter )
         parameter_ = zeroOrMoreParameter;
   }
   return *this;
}


inline OptionDescription &
OptionDescription::setStorage( const ValueBinder &storage )
{
   storage_ = storage;
   return *this;
}


inline const Value &
OptionDescription::defaultValue() const
{
   return defaultValue_;
}


inline bool 
OptionDescription::hasDefaultValue() const
{
   return !defaultValue_.isEmpty();
}


inline bool 
OptionDescription::hasStorage() const
{
   return !storage_.isEmpty();
}


inline bool 
OptionDescription::hasShortName() const
{
   return !shortName_.empty();
}


inline bool 
OptionDescription::hasLongName() const
{
   return !longName_.empty();
}


inline String 
OptionDescription::shortName() const
{
   return shortName_;
}


inline String 
OptionDescription::longName() const
{
   return longName_;
}


inline String 
OptionDescription::name() const
{
   return longName_.empty() ? "-" + shortName() : longName_;
}


inline String 
OptionDescription::description() const
{
   return description_;
}


inline bool 
OptionDescription::hasNoParameter() const
{
   return parameter_ == noParameter;
}


inline bool 
OptionDescription::mayHaveParameter() const
{
   return parameter_ == optionalParameter 
          ||  parameter_ == zeroOrMoreParameter;
}


inline bool 
OptionDescription::requireParameter() const
{
   return parameter_ == requiredParameter
          ||  parameter_ == oneOrMoreParameter;
}


inline Value
OptionDescription::validate( const Parameters &parameters ) const
{
   if ( requireParameter()  &&  parameters.empty() )
      throw ValidatorError( ValidatorError::missingParameter, "required parameter is missing." );

   if ( requireParameter()
        ||  ( mayHaveParameter()  &&  !parameters.empty() ) )
   {
      if ( !validator_ )
         throw ValidatorError( ValidatorError::undefinedValidator, "no validator set for this option." );
      return validator_->validate( parameters );
   }

   return defaultValue();
}


inline void
OptionDescription::store( const Value &value ) const
{
   if ( hasStorage() )
      storage_.set( value );
}



// class OptionsDescription
// //////////////////////////////////////////////////////////////////
inline
OptionsDescription::OptionsDescription( const String &caption )
   : caption_( caption )
{
}


inline
OptionsDescription::~OptionsDescription()
{
}


inline OptionsDescription &
OptionsDescription::add( const OptionDescription &description )
{
   int index = options_.size();
   options_.push_back( description );

   if ( longNameIndexes_.count( description.longName() ) > 0 )
      throw OptionsDescriptionError( "More than one option registered with long name '" 
                                     + description.longName() + "'." );
   if ( shortNameIndexes_.count( description.shortName() ) > 0 )
      throw OptionsDescriptionError( "More than one option registered with short name '" 
                                     + description.shortName() + "'." );

   if ( description.hasLongName() )
   {
      longNameIndexes_.insert( OptionIndexes::value_type( description.longName(), index ) );
      nameIndexes_.insert( OptionIndexes::value_type( description.longName(), index ) );
   }

   if ( description.hasShortName() )
   {
      shortNameIndexes_.insert( OptionIndexes::value_type( description.shortName(), index ) );
      nameIndexes_.insert( OptionIndexes::value_type( "-" + description.shortName(), index ) );
   }

   return *this;
}


inline OptionsDescription &
OptionsDescription::add( const String &optionName,
                         const String &description )
{
   return add( option<std::string>( optionName, description ) );
}


inline std::set<String>
OptionsDescription::optionNames() const
{
   std::set<String> optionNames;
   Options::const_iterator it = options_.begin();
   for ( ; it != options_.end(); ++it )
   {
      const OptionDescription &option = *it;
      optionNames.insert( option.name() );
   }
   return optionNames;
}


inline bool 
OptionsDescription::hasOption( const String &optionName ) const
{
   return nameIndexes_.count( optionName) > 0;
}


inline const OptionDescription &
OptionsDescription::description( const String &optionName ) const
{
   OptionIndexes::const_iterator it = nameIndexes_.find( optionName );
   if ( it == nameIndexes_.end() )
   {
      std::ostringstream os;
      os << "No option registered for name '" + optionName + "'.";
      throw OptionsDescriptionError( os.str() );
   }

   return options_[it->second];
}

// template function option
// //////////////////////////////////////////////////////////////////

template<class ValueType>
OptionDescription option( const std::string &optionName,
                          const std::string &description )
{
   return OptionDescription( optionName, description, defaultValidator<ValueType>() );
}

template<class DefaultValueType>
OptionDescription optionDefault( const std::string &optionName,
                                 const DefaultValueType &defaultValue,
                                 const std::string &description )
{
   Value value;
   set( value, defaultValue );
   OptionDescription option( optionName, description, defaultValidator<DefaultValueType>() );
   return option.setDefaultValue( value );
}

template<class StorageValueType>
OptionDescription option( const std::string &optionName,
                          StorageValueType &storageValue,
                          const std::string &description )
{
   OptionDescription option( optionName, description, defaultValidator<StorageValueType>() );
   return option.setStorage( valueBinder( storageValue ) );
}

template<class StorageValueType
        ,class DefaultValueType>
OptionDescription option( const std::string &optionName,
                          StorageValueType &storageValue,
                          const DefaultValueType &defaultValue,
                          const std::string &description )
{
   Value value;
   set( value, StorageValueType( defaultValue ) );
   OptionDescription option( optionName, description, defaultValidator<StorageValueType>() );
   option.setDefaultValue( value );
   return option.setStorage( valueBinder( storageValue ) );
}


template<class ValueType>
OptionDescription validatedOption( const std::string &optionName,
                                   const std::string &description,
                                   const ValidatorPtr &validator )
{
   return OptionDescription( optionName, description, validator );
}

template<class DefaultValueType>
OptionDescription validatedOptionDefault( const std::string &optionName,
                                          const DefaultValueType &defaultValue,
                                          const std::string &description,
                                          const ValidatorPtr &validator )
{
   Value value;
   set( value, defaultValue );
   OptionDescription option( optionName, description, validator );
   return option.setDefaultValue( value );
}

template<class StorageValueType>
OptionDescription validatedOption( const std::string &optionName,
                                   StorageValueType &storageValue,
                                   const std::string &description,
                                   const ValidatorPtr &validator )
{
   OptionDescription option( optionName, description, validator );
   return option.setStorage( valueBinder( storageValue ) );
}

template<class StorageValueType
        ,class DefaultValueType>
OptionDescription validatedOption( const std::string &optionName,
                                   StorageValueType &storageValue,
                                   const DefaultValueType &defaultValue,
                                   const std::string &description,
                                   const ValidatorPtr &validator )
{
   Value value;
   set( value, StorageValueType( defaultValue ) );
   OptionDescription option( optionName, description, validator );
   option.setDefaultValue( value );
   return option.setStorage( valueBinder( storageValue ) );
}



} } // namespace CommandLineOptions

#endif // CPPUTTOOLS_CONFIGURATION_DESCRIPTION_H_INCUDED
