#ifndef CPPUTTOOLS_CONFIGURATION_VALIDATOR_H_INCUDED
# define CPPUTTOOLS_CONFIGURATION_VALIDATOR_H_INCUDED

# include "errors.h"
# include <cpputtools/value.h>  // <cpputtools/forwards.h>

namespace CppUTTools { namespace CommandLineOptions {


class Validator : public ReferenceCountedPolicy
{
public:
   virtual ~Validator() {}

   virtual Value validate( const Parameters &parameters ) const = 0;
};


template<class OptionType>
class StringStreamValidator : public Validator
{
public: // overridden from Validator
   Value validate( const Parameters &parameters ) const;
};

template<class OptionType>
ValidatorPtr defaultValidator();

template<class OptionType>
class EnumerationValidator : public Validator
{
public:
   void add( const String &text, 
             const OptionType &value,
             const String &help = "" );

   std::deque<String> inputs() const;

   String inputHelp( const String &text ) const;

   ValidatorPtr clone() const;

public: // overridden from Validator
   Value validate( const Parameters &parameters ) const;

private:
   struct ValueData
   {
      OptionType value_;
      String help_;
   };

   typedef std::map<String,ValueData> Values;
   Values values_;
   std::deque<String> inputs_;
};




template<class Functor>
ValidatorPtr makeValidator( Functor functor );

template<class OptionType>
std::string validatorAdditionalMessage();

#undef COMMANDLINEOPTIONS_VALIDATOR_MESSAGE
#define COMMANDLINEOPTIONS_VALIDATOR_MESSAGE( OptionType, message )  \
   template<>                                                        \
   inline std::string validatorAdditionalMessage<OptionType>()       \
   {                                                                 \
      return message;                                                \
   }

COMMANDLINEOPTIONS_VALIDATOR_MESSAGE( int, "is not an integer." )
COMMANDLINEOPTIONS_VALIDATOR_MESSAGE( unsigned int, "is not an integer." )
COMMANDLINEOPTIONS_VALIDATOR_MESSAGE( short, "is not an integer." )
COMMANDLINEOPTIONS_VALIDATOR_MESSAGE( unsigned short, "is not an integer." )
COMMANDLINEOPTIONS_VALIDATOR_MESSAGE( long, "is not an integer." )
COMMANDLINEOPTIONS_VALIDATOR_MESSAGE( unsigned long, "is not an integer." )
COMMANDLINEOPTIONS_VALIDATOR_MESSAGE( float, "is not a real number." )
COMMANDLINEOPTIONS_VALIDATOR_MESSAGE( double, "is not a real number." )
COMMANDLINEOPTIONS_VALIDATOR_MESSAGE( long double, "is not a real number." )
COMMANDLINEOPTIONS_VALIDATOR_MESSAGE( char, "is not a character." )
COMMANDLINEOPTIONS_VALIDATOR_MESSAGE( std::string, "is not a string." )
//COMMANDLINEOPTIONS_VALIDATOR_MESSAGE( wchar_t, "is not a character." )


// Template StringStreamValidator
// //////////////////////////////////////////////////////////////////

template<class OptionType>
Value
StringStreamValidator<OptionType>::validate( const Parameters &parameters ) const
{
   IStringStream is( parameters.at(0) );
   OptionType optionValue;
   is >> optionValue;
   is >> std::ws;
   if ( !is.eof() )  // input not completly consumed
   {
      std::ostringstream os;
      os  <<  "parameter '" << parameters.at(0) << "', ";
      os  <<  validatorAdditionalMessage<OptionType>();
      throw ValidatorError( ValidatorError::badParameterValue, os.str() );
   }
   return value( optionValue );
}


template<class OptionType>
ValidatorPtr 
defaultValidator()
{
   return ValidatorPtr( new StringStreamValidator<OptionType>() );
}


// Template EnumerationValidator
// //////////////////////////////////////////////////////////////////


template<class OptionType>
ValidatorPtr
EnumerationValidator<OptionType>::clone() const
{
   return ValidatorPtr( new EnumerationValidator<OptionType>( *this ) );
}


template<class OptionType>
void
EnumerationValidator<OptionType>::add( const String &text, 
                                       const OptionType &value,
                                       const String &help )
{
   ValueData data;
   data.value_ = value;
   data.help_ = help;
   values_.insert( Values::value_type( text, data ) );
   inputs_.push_back( text );
}


template<class OptionType>
std::deque<String> 
EnumerationValidator<OptionType>::inputs() const
{
   return inputs_;
}


template<class OptionType>
String 
EnumerationValidator<OptionType>::inputHelp( const String &text ) const
{
   CPPTL_TYPENAME Values::const_iterator it = values_.find( text );
   if ( it == values_.end() )
      return "";
   return it->second.help_;
}


template<class OptionType>
Value
EnumerationValidator<OptionType>::validate( const Parameters &parameters ) const
{
   String text = parameters.at(0);
   CPPTL_TYPENAME Values::const_iterator it = values_.find( text );
   if ( it == values_.end() )
      throw ValidatorError( ValidatorError::badParameterValue, "value must be..." );
   return value( it->second.value_ );
}

// Function makeValidator
// //////////////////////////////////////////////////////////////////

template<class Functor>
class ValidatorImpl : public Validator
{
public:
   ValidatorImpl( Functor functor )
      : functor_( functor )
   {
   }

   // overridden from Validator
   Value validate( const Parameters &parameters ) const
   {
      return functor_( parameters );
   }

private:
   Functor functor_;
};

template<class Functor>
ValidatorPtr validator( Functor functor )
{
   return ValidatorPtr( new ValidatorImpl<Functor>( functor ) );
}


} } // namespace CommandLineOptions

#endif // CPPUTTOOLS_CONFIGURATION_VALIDATOR_H_INCUDED
