#include <cpput/ipc.h>
#include <json/json.h>

namespace CppUT {

// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////
// class SynchronousWriteChannel
// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////

SynchronousWriteChannel::~SynchronousWriteChannel()
{
}


void 
SynchronousWriteChannel::close()
{
   closeRead();
   closeWrite();
}



// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////
// class JsonRPCRequest
// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////

JsonRPCRequest::JsonRPCRequest( const Json::Value &requestId )
   : id_( requestId )
{
}

const Json::Value &
JsonRPCRequest::requestId() const
{
   return id_;
}



// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////
// class JsonRPCRequestHandler
// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////

JsonRPCRequestHandler::~JsonRPCRequestHandler()
{
}



// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////
// class JsonRPCRequestHandler
// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////

JsonRPCError JsonRPCError::parseError( -32700, "Parse error." );
JsonRPCError JsonRPCError::invalidRequest( -32600, "Invalid Request." );
JsonRPCError JsonRPCError::methodNotFound( -32601, "Method not found." );
JsonRPCError JsonRPCError::invalidParameter( -32602, "Invalid params." );
JsonRPCError JsonRPCError::internalError( -32603, "Internal error." );


JsonRPCError::JsonRPCError( int errorCode, const char *message )
   : errorCode_( errorCode )
   , message_( message )
{
}


JsonRPCError::JsonRPCError( int errorCode, const char *message, const Json::Value &detail )
   : errorCode_( errorCode )
   , message_( message )
   , detail_( detail )
{
}


JsonRPCError 
JsonRPCError::detail( const Json::Value &detail )
{
   return JsonRPCError( errorCode_, message_.c_str(), detail );
}


int 
JsonRPCError::errorCode() const
{
   return errorCode_;
}


const char *
JsonRPCError::message() const
{
   return message_.c_str();
}


const Json::Value &
JsonRPCError::detail() const
{
   return detail_;
}



// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////
// class JsonRPCChannel
// //////////////////////////////////////////////////////////////////
// //////////////////////////////////////////////////////////////////

JsonRPCChannel::JsonRPCChannel( SynchronousWriteChannelAutoPtr channel,
                                JsonRPCRequestHandler &requestHandler )
   : channel_( channel )
   , requestHandler_( requestHandler )
   , stopWaiting_( false )
   , nextRequestId_( 0 )
{
   CPPTL_ASSERT_MESSAGE( channel_.get() != 0, "A valid channel must be provided" );
}


JsonRPCChannel::~JsonRPCChannel()
{
}


void 
JsonRPCChannel::close()
{
   channel_->close();
   stopWaitingForRequests();
}


void 
JsonRPCChannel::waitForRequests()
{
   std::vector<char> buffer( 16384 );
   while ( !stopWaiting_ )
   {
      size_t readLength = 0;
      if ( !channel_->read( &buffer[0], buffer.size(), readLength ) )
      {
         break;
      }
      decodeIncommingData( &buffer[0], readLength );
      processIncommingData();
   }

   stopWaiting_ = false;
}


void 
JsonRPCChannel::stopWaitingForRequests()
{
   stopWaiting_ = true;
}


void 
JsonRPCChannel::sendRequest( const char *method, const Json::Value &parameter )
{
   Json::Value request;
   request["jsonrpc"] = "2.0";
   request["method"] = method;
   if ( !parameter.isNull() )
   {
      CPPTL_ASSERT_MESSAGE( parameter.isObject(), "JSON RPC parameter must be object" );
      request["params"] = parameter;
   }
   ++nextRequestId_;
   request["id"] = nextRequestId_;
   sendMessage( request );
}


void 
JsonRPCChannel::sendNotification( const char *method, const Json::Value &parameter )
{
   Json::Value request;
   request["jsonrpc"] = "2.0";
   request["method"] = method;
   if ( !parameter.isNull() )
   {
      CPPTL_ASSERT_MESSAGE( parameter.isObject(), "JSON RPC parameter must be object" );
      request["params"] = parameter;
   }
   sendMessage( request );
}


void 
JsonRPCChannel::reply( JsonRPCRequest request,
                       const Json::Value &result )
{
   Json::Value reply;
   reply["jsonrpc"] = "2.0";
   reply["result"] = result;
   reply["id"] = request.requestId();
   sendMessage( reply );
}


void 
JsonRPCChannel::replyWithError( JsonRPCRequest request,
                                const JsonRPCError &error )
{
   Json::Value reply;
   reply["jsonrpc"] = "2.0";
   Json::Value errorValue;
   errorValue["code"] = error.errorCode();
   errorValue["message"] = error.message();
   if ( !error.detail().isNull() )
      errorValue["data"] = error.detail();
   reply["error"] = errorValue;
   reply["id"] = request.requestId();
   sendMessage( reply );
}


void 
JsonRPCChannel::sendMessage( const Json::Value &message )
{
   Json::FastWriter writer;
   
   std::string messageAsString = writer.write( message );
   messageAsString += "\n"; // RPC Message end tag

   channel_->write( messageAsString.c_str(), messageAsString.length() );
}


void 
JsonRPCChannel::decodeIncommingData( const char *begin, size_t incommingLength )
{
   incommingData_.append( begin, incommingLength );

   std::string::size_type pos = 0;
   std::string::size_type nextEOL = std::string::npos;
   while (true)
   {
      nextEOL = incommingData_.find( '\n', pos );
      if ( nextEOL == std::string::npos )
      {
         break;
      }
      std::string line = incommingData_.substr( pos, nextEOL - pos );
      if ( !line.empty()  &&  line[line.size()-1] == '\r' )
      {
         line.erase( line.size()-1 );
      }
      pendingRequests_.push_back( line );

      pos = nextEOL + 1;
   }
   std::string newInput = incommingData_.substr( pos );
   incommingData_ = newInput;
}


// @todo handling of reply
void 
JsonRPCChannel::processIncommingData()
{
   while ( !stopWaiting_  &&  !pendingRequests_.empty() )
   {
      Json::Value request;
      {
         Json::Reader reader;
         const CppTL::ConstString &message = pendingRequests_.front();
         if ( !reader.parse( message.c_str(), message.c_str() + message.length(), request, false ) )
         {
            // log malformatted message.
            fprintf( stderr, "JSONRPC: invalid incomming message" );
            // Should send a notification ?
         }
         pendingRequests_.pop_front();
      }
      if ( request.isObject() )
      {
         JsonRPCRequest requestId( request.get("id", Json::Value::null) );
         const Json::Value &method = request.get("method", Json::Value::null);
         if ( !method.isString()  ||  requestId.requestId().isArray()  ||  requestId.requestId().isObject() )
         {
            replyWithError( requestId, JsonRPCError::invalidRequest );
         }
         else if ( requestId.requestId().isNull() ) // Notification
         {
            requestHandler_.onNotification( method.asString(), 
                                            request.get("params", Json::Value::null),
                                            *this );
            
         }
         else // Request
         {
            requestHandler_.onRequest( requestId, 
                                       method.asString(), 
                                       request.get("params", Json::Value::null),
                                       *this );
         }
      }
   }
}


} // namespace CppUT {