#if defined(_WIN32)
# include <cpput/ipc.h>
# include <io.h>
# include <stdio.h>

# define WIN32_LEAN_AND_MEAN 
# define NOGDI
# define NOUSER
# define NOKERNEL
# define NOSOUND
# define NOMINMAX
# ifndef BLENDFUNCTION
#  define BLENDFUNCTION void    /* for mingw & gcc */
# endif
# include <windows.h> 

namespace CppUT {
   namespace Impl {

      class Win32SynchronousWriteChannel : public SynchronousWriteChannel
      {
      public:
         Win32SynchronousWriteChannel( HANDLE readHandle, HANDLE writeHandle )
            : readHandle_( readHandle )
            , writeHandle_( writeHandle )
         {
         }

         virtual ~Win32SynchronousWriteChannel()
         {
            close();
         }

      public: // overridden from SynchronousWriteChannel

         virtual bool read( void *readBuffer, size_t bufferSize, size_t &readLength )
         {
            DWORD readBytes = 0;
            BOOL ok = ReadFile( readHandle_, readBuffer, bufferSize, &readBytes, NULL );
            if ( ok )
            {
               readLength = static_cast<size_t>( readBytes );
               //fprintf( stderr, "Win32Read: %*s\n", readLength, readBuffer );
               return true;
            }
            DWORD error = GetLastError();
            if ( error == ERROR_HANDLE_EOF  
                 ||  error == ERROR_OPERATION_ABORTED ) // Ctrl+C, only if StdIn is Console.
            {
               readLength = 0;
               return true;
            }
            // If the pipe is in message mode: ERROR_MORE_DATA
            // See Also: PeekNamedPipe 
            return false;
         }

         virtual bool write( const void *buffer, size_t length )
         {
            if ( writeHandle_ == INVALID_HANDLE_VALUE )
            {
               return false;
            }
            DWORD written = 0;
            BOOL ok = WriteFile( writeHandle_, 
                                 buffer, static_cast<DWORD>(length), 
                                 &written, NULL );
            if (!ok )
            {
               DWORD error = GetLastError();
               fprintf( stderr, "Failed to write to stdout, error code: %d\n", error );
            }
            else
            {
               //fprintf( stderr, "Win32Write: %*s\n", length, buffer );
            }
            return ok  && written == length;
         }

         virtual void closeRead()
         {
            if ( readHandle_ != INVALID_HANDLE_VALUE )
            {
               CloseHandle( readHandle_ );
               readHandle_ = INVALID_HANDLE_VALUE;
            }
         }

         virtual void closeWrite()
         {
            if ( writeHandle_ != INVALID_HANDLE_VALUE )
            {
               CloseHandle( writeHandle_ );
               writeHandle_ = INVALID_HANDLE_VALUE;
            }
         }

      private:
         HANDLE readHandle_;
         HANDLE writeHandle_;
      };

   } // namespace Impl {


// ReadFile() to read or WriteFile()
// CloseHandler() to close, DuplicateHandle()
// CreatePipe() to create
// GetStdHandle

SynchronousWriteChannelAutoPtr
makeSlaveChannel()
{
   HANDLE stdinHandle = GetStdHandle( STD_INPUT_HANDLE );
   HANDLE stdoutHandle = GetStdHandle( STD_OUTPUT_HANDLE );

   if ( stdinHandle == INVALID_HANDLE_VALUE
        ||  stdoutHandle == INVALID_HANDLE_VALUE ) 
   {
      // Error, can not bind stdin or stdout.
      return SynchronousWriteChannelAutoPtr();
   }

   SynchronousWriteChannelAutoPtr channel( new Impl::Win32SynchronousWriteChannel( 
      stdinHandle, stdoutHandle ) );

   // Redirect stdout to stderr (so that printf to not corrupt the pipe)
//   _dup2( 2, 1 );

   return channel;
}



} // namespace CppUT {


#endif // #if defined(_WIN32)
