#include <cpput/runner.h>
#include <cpput/lighttestrunner.h>
#include <cpput/ipc.h>
#include <algorithm>
#include <deque>
#include <map>
#include <string.h>

namespace CppUT {
   namespace Impl {
      // ////////////////////////////////////////////////////////////
      // ////////////////////////////////////////////////////////////
      // class Impl::RunnerImpl
      // ////////////////////////////////////////////////////////////
      // ////////////////////////////////////////////////////////////
      class RunnerImpl
      {
      public:
         RunnerImpl( int argc, const char *argv[] )
            : isMonitoredSlave_( false )
         {
            std::string slaveFlag = "--test-slave";
            for ( int index = 0; index < argc; ++index )
            {
               if ( argv[index] == slaveFlag )
               {
                  isMonitoredSlave_ = true;
                  memcpy( &argv[index], &argv[index+1], sizeof(argv[0]) * (argc-index-1) );
                  argv[argc] = 0;
                  --argc;
               }
            }
         }

         virtual ~RunnerImpl()
         {
         }

         bool isMonitoredSlave_;
      };

      // ////////////////////////////////////////////////////////////
      // ////////////////////////////////////////////////////////////
      // class Impl::MonitoredSlaveTestResultUpdater
      // ////////////////////////////////////////////////////////////
      // ////////////////////////////////////////////////////////////
      
      class MonitoredSlaveTestResultUpdater : public TestResultUpdater
      {
      public:
         MonitoredSlaveTestResultUpdater( JsonRPCChannel &channel )
            : channel_( channel )
            , testId_( -1 )
         {
         }

         virtual ~MonitoredSlaveTestResultUpdater()
         {
         }

         void beginNewTest( int testId )
         {
            testId_ = testId;
            events_.clear();
         }

         const Json::Value &events() const
         {
            return events_;
         }

      public: // overridden from TestResultUpdater
         virtual void addResultLog( const Json::Value &log )
         {
            Json::Value eventObject;
            eventObject["type"] = "log";
            eventObject["data"] = log;
            events_.append( eventObject );
         }

         virtual void addResultAssertion( const Assertion &assertion )
         {
            Json::Value eventObject;
            eventObject["type"] = "assertion";
            eventObject["kind"] = assertion.kind() == Assertion::fault ? "fault"
                                                                       : "assertion";
            if ( assertion.location().isValid() )
            {
               Json::Value location;
               location["source"] = assertion.location().file_;
               location["line"] = assertion.location().line_;
               eventObject["location"] = location;
            }
            if ( assertion.isIgnoredFailure() )
            {
               eventObject["ignoredFailure"] = true;
            }
            eventObject["detail"] = assertion.detail();
            events_.append( eventObject );
         }

      private:
         JsonRPCChannel &channel_;
         Json::Value events_;
         int testId_;
      };

      // ////////////////////////////////////////////////////////////
      // ////////////////////////////////////////////////////////////
      // class Impl::MonitoredSlaveRunner
      // ////////////////////////////////////////////////////////////
      // ////////////////////////////////////////////////////////////
      

      class MonitoredSlaveRunner : private JsonRPCRequestHandler
      {
      public:
         virtual ~MonitoredSlaveRunner()
         {
         }

         bool run();

      private: // overridden from JsonRPCRequestHandler
         virtual void onRequest( JsonRPCRequest request,
                                 const std::string &method,
                                 const Json::Value &parameter,
                                 JsonRPCChannel &channel );

         virtual void onNotification( const std::string &method, 
                                      const Json::Value &parameter, 
                                      JsonRPCChannel &channel );

      private:
         void onListTestsRequest( JsonRPCRequest request, 
                                  JsonRPCChannel &channel );
         void onRunTestsRequest( JsonRPCRequest request, 
                                 const Json::Value &parameter,
                                 JsonRPCChannel &channel );
         void onTerminateRequest( JsonRPCRequest request, 
                                  JsonRPCChannel &channel );

      private:
         typedef std::deque<const TestMeta *> Tests;
         // Retrieves or create the id for the specified test suite
         int getSuiteId( Suite suite );
         // Retrieves or create the id for the specified test case
         int getTestId( const TestMeta *test );
         void runMonitoredTests( const Tests &tests, 
                                 JsonRPCChannel &channel );

         typedef std::map<int, Suite> SuitesById;
         SuitesById suitesById_;
         typedef std::map<Suite, int> IdsBySuite;
         IdsBySuite idsBySuite_;
         typedef std::map<int, const TestMeta *> TestsById;
         TestsById testsById_;
         typedef std::map<const TestMeta *, int> IdsByTest;
         IdsByTest idsByTest_;

      };


      bool
      MonitoredSlaveRunner::run()
      {
         CppUT::SynchronousWriteChannelAutoPtr transportChannel = CppUT::makeSlaveChannel();
         JsonRPCChannel channel( transportChannel, *this );
         channel.waitForRequests();
         return true;
      }


      void 
      MonitoredSlaveRunner::onRequest( JsonRPCRequest request,
                                       const std::string &method,
                                       const Json::Value &parameter,
                                       JsonRPCChannel &channel )
      {
         if ( method == "listTests" )
            onListTestsRequest( request, channel );
         else if ( method == "runTests" )
            onRunTestsRequest( request, parameter, channel );
         else if ( method == "terminate" )
            onTerminateRequest( request, channel );
      }


      void 
      MonitoredSlaveRunner::onNotification( const std::string &method, 
                                            const Json::Value &parameter, 
                                            JsonRPCChannel &channel )
      {
         // No notification message in monitored slave
      }

      int 
      MonitoredSlaveRunner::getSuiteId( Suite suite )
      {
         IdsBySuite::iterator it = idsBySuite_.find( suite );
         if ( it != idsBySuite_.end() )
            return it->second;
         int id = suitesById_.size() + 1;
         suitesById_[id] = suite;
         idsBySuite_[suite] = id;
         return id;
      }

      int 
      MonitoredSlaveRunner::getTestId( const TestMeta *test )
      {
         IdsByTest:: iterator it = idsByTest_.find( test );
         if ( it != idsByTest_.end() )
            return it->second;
         int id = testsById_.size() + 1;
         testsById_[id] = test;
         idsByTest_[test] = id;
         return id;
      }

      void 
      MonitoredSlaveRunner::onListTestsRequest( JsonRPCRequest request, 
                                                JsonRPCChannel &channel )
      {
         Json::Value result( Json::arrayValue );
         std::deque<Suite> suitesToVisit;
         suitesToVisit.push_back( CppUT::Registry::getRootSuite() );
         while ( !suitesToVisit.empty() )
         {
            Suite suite = suitesToVisit.front();
            suitesToVisit.pop_front();
            // Adds suite detail to result
            Json::Value suiteObject;
            int suiteId = getSuiteId( suite );
            suiteObject["type"] = "suite";
            suiteObject["id"] = suiteId;
            suiteObject["name"] = suite.name();
            Json::Value parent;
            if ( suite.parent().isValid() )
               parent = getSuiteId( suite.parent() );
            suiteObject["parent"] = parent;
            result.append( suiteObject );
            // Explore child suites
            int nbChildSuite = suite.nestedSuiteCount();
            for ( int suiteIndex = 0; suiteIndex < nbChildSuite; ++suiteIndex )
            {
               Suite childSuite = suite.nestedSuiteAt( suiteIndex );
               suitesToVisit.push_back( childSuite );
            }
            // Explore contained test cases
            int nbTestCase = suite.testCaseCount();
            for ( int testIndex =0; testIndex < nbTestCase; ++testIndex )
            {
               const TestMeta *test = suite.testCaseAt( testIndex );
               Json::Value testObject;
               testObject["type"] = "test";
               testObject["id"] = getTestId( test );
               testObject["name"] = test->name();
               testObject["parent"] = suiteId;
               result.append( testObject );
            }
         }
         channel.reply( request, result );
      }


      void 
      MonitoredSlaveRunner::onRunTestsRequest( JsonRPCRequest request, 
                                               const Json::Value &parameter,
                                               JsonRPCChannel &channel )
      {
         Tests tests;
         bool ok = parameter.isMember( "tests" );
         const char *errorDetail = "'tests' parameter is missing";
         if ( ok )
         {
            Json::Value::const_iterator it = parameter["tests"].begin();
            Json::Value::const_iterator itEnd = parameter["tests"].end();
            for ( ; it != itEnd; ++it )
            {
               const Json::Value &value = *it;
               if ( !value.isInt() )
               {
                  errorDetail = "Non integer id passed in 'tests'";
                  ok = false;
                  break;
               }
               TestsById::iterator itTest = testsById_.find(value.asInt());
               if ( itTest == testsById_.end() )
               {
                  errorDetail = "Invalid test id passed in 'tests'";
                  ok = false;
                  break;
               }
               tests.push_back( itTest->second );
            }
         }
         if ( !ok )
         {
            channel.replyWithError( request, JsonRPCError::invalidParameter.detail(errorDetail) );
         }
         else
         {
            runMonitoredTests( tests, channel );

            channel.reply( request, Json::nullValue );
         }
      }


      void 
      MonitoredSlaveRunner::onTerminateRequest( JsonRPCRequest request, 
                                                JsonRPCChannel &channel )
      {
         channel.stopWaitingForRequests();
      }


      void 
      MonitoredSlaveRunner::runMonitoredTests( const Tests &tests, 
                                               JsonRPCChannel &channel )
      {
         MonitoredSlaveTestResultUpdater updater( channel );
         TestInfo::threadInstance().setTestResultUpdater( updater );
         Tests::const_iterator itTestEnd = tests.end();
         for ( Tests::const_iterator itTest = tests.begin(); itTest != itTestEnd; ++itTest )
         {
            const TestMeta *test = *itTest;
            int testId = getTestId( test );

            updater.beginNewTest( testId );
            Json::Value testStartedNotification;
            testStartedNotification["id"] = testId;
            channel.sendNotification( "testStarted", testStartedNotification );

            test->runTest();
            
            TestStatus &testStatus = TestInfo::threadInstance().testStatus();
            Json::Value statusObject;

            const char *statusAsText;
            switch ( testStatus.status() )
            {
               case TestStatus::passed: 
                  statusAsText = "passed"; 
                  break;
               case TestStatus::skipped: 
                  statusAsText = "skipped"; 
                  break;
               case TestStatus::failed: 
                  statusAsText = "failed"; 
                  break;
               default: 
                  CPPTL_DEBUG_ASSERT_UNREACHABLE;
                  statusAsText = "?"; 
                  break;
            }
            statusObject["status"] = statusAsText;
            Json::Value statistics;
            statistics["assertions"] = testStatus.assertionCount();
            statistics["failedAssertion"] = testStatus.failedAssertionCount();
            statistics["ignoredFailures"] = testStatus.ignoredFailureCount();
            statusObject["statistics"] = statistics;

            Json::Value testEndedNotification;
            testEndedNotification["id"] = testId;
            testEndedNotification["events"] = updater.events();
            testEndedNotification["status"] = statusObject;
            channel.sendNotification( "testEnded", testEndedNotification );
         }
      }


   } // namespace Impl {



   Runner::Runner( int argc, const char *argv[] )
      : impl_( new Impl::RunnerImpl( argc, argv ) )
   {
   }

   Runner::Runner( int argc, char *argv[] )
      : impl_( new Impl::RunnerImpl( argc, const_cast<const char **>(argv) ) )
   {
   }

   Runner::~Runner()
   {
      delete impl_;
   }

   bool 
   Runner::run()
   {
      if ( impl_->isMonitoredSlave_ )
      {
         //_asm int 3;
         Impl::MonitoredSlaveRunner runner;
         return runner.run();
      }
      else
      {
         LightTestRunner runner;
         runner.addSuite( CppUT::Registry::getRootSuite() );
         return runner.runTests();
      }
   }



} // namespace CppUT {