import subprocess
import json
import types
import logging
import traceback

# TODO: handle IOError

class PipeTransport:
    """Transport that provides synchronous write operation to guaranty delivery.
    """
    def __init__( self, inFile, outFile,
                  my_name ):
        self.inFile = inFile
        self.outFile = outFile
        self.name = my_name
        self.logger = logging.getLogger('PipeTransport')

    def read_line( self ):
        line = ''
        while not line:
            line = self.inFile.readline()
            if not line: # empty line means EOF was reached, e.g. pipe is broken
                self.logger.debug( 'Read pipe was closed by peer' )
                raise EOFError( 'Read pipe was closed by peer' )
            line = line.strip()
        self.logger.debug( 'Received: %s', line )
        return line

    def write_line( self, line ):
        self.logger.debug( 'Sending: %s', line )
        self.outFile.write( line + '\n' )

    def close_read( self ):
        self.logger.debug( 'Closing read' )
        self.inFile.close()

    def close_write( self ):
        self.logger.debug( 'Closing write' )
        self.outFile.close()

    def close( self ):
        self.close_read()
        self.close_write()

class SubprocessPipeTransport:
    def __init__( self, *args, **kwargs ):
        """Open a connection with a new sub-process.
           Parameters: same as subprocess.Popen.
           """
        kwargs['stdin'] = subprocess.PIPE
        kwargs['stdout'] = subprocess.PIPE
        self.process = subprocess.Popen( *args, **kwargs )
        self._pipe = PipeTransport( self.process.stdout, self.process.stdin, 'Runner' )
        self.logger = logging.getLogger('SubprocessPipeTransport')

    def read_line( self ):
        return self._pipe.read_line()

    def write_line( self, buffer ):
        self._pipe.write_line( buffer )

    def close( self ):
        """Kill the sub-process."""
        self._pipe.close()
        if self.process.poll() is None: # process has not terminated yet
            try:
                self.process.kill()
            except WindowsError, e:
                self.logger.debug( 'Error while killing sub-process.', exc_info = True )

class JsonRPCError(IOError):
    def __init__( self, code, message, detail = None ):
        if detail and not isinstance(detail, dict):
            detail = str(detail)
        self.data = {'code': code, 'message': message }
        if detail:
            self.data['detail'] = detail
            as_string = 'code[%d]: %s, %s' % (code, message, detail)
        else:
            as_string = 'code[%d]: %s' % (code, message)
        IOError.__init__( self, as_string )

    def has_same_code( self, other ):
        return self.data['code'] == other.data['code']

    def detail( self, detail ):
        return JsonRPCError( self.data['code'], self.data['message'], detail )

    def detail_from_exception( self, exception ):
        return self.detail( {'exception': exception.__class__.__name__,
                             'detail': str(exception)} )

PARSE_ERROR = JsonRPCError( -32700, "Parse error." )
INVALID_REQUEST = JsonRPCError( -32600, "Invalid Request." )
METHOD_NOT_FOUND = JsonRPCError( -32601, "Method not found." )
INVALID_PARAMETER = JsonRPCError( -32602, "Invalid params." )
INTERNAL_ERROR = JsonRPCError( -32603, "Internal error." )
INVALID_REPLY = JsonRPCError( -32604, "Invalid reply." )
CONNECTION_LOST = JsonRPCError( -32605, "Connection lost." )

class RequestHandler:
    def __init__( self ):
        self._handlers = {} # {method: (callable, is_async)}

    def register_synchronous_method( self, method, handler ):
        """Register the callable handler to handle RPC method.
           Reply with the result of the handler as a JSON object.
           If the handler raise an exception, it is automatically caught and request is replied to with an error.
           handler: will be called with (parameter, channel)
        """
        self._handlers[method] = (handler, False)

    def register_asynchronous_method( self, method, handler ):
        """Register the callable handler to handle RPC method.
           If the handler raise an exception, it is automatically caught and request is replied to with an error.
           It is up to the user to actually reply to the request.
           handler: will be called with (request_id, parameter, channel)
        """
        self._handlers[method] = (handler, True)

    def register_notification_method( self, method, handler ):
        """Register the callable handler to handle RPC method.
           If the handler raise an exception, it is automatically caught.
           handler: will be called with (parameter, channel)
        """
        self._handlers[method] = (handler, True)
    
    def on_notification( self, method, parameter, channel ):
        if method not in self._handlers:
            channel.reply_with_error( None, METHOD_NOT_FOUND.detail(method) )
        try:
            handler, is_async = self._handlers[method]
            assert is_async
            handler( parameter, channel )
        except Exception, e:
            channel.reply_with_exception( None, e )

    def onRequest( self, request_id, method, parameter, channel ):
        if method not in self._handlers:
            channel.reply_with_error( request_id, METHOD_NOT_FOUND.detail(method) )
        try:
            handler, is_async = self._handlers[method]
            if is_async:
                handler( request_id, parameter, channel )
            else:
                result = handler( parameter, channel )
                channel.reply( request_id, result )
        except Exception, e:
            channel.reply_with_exception( request_id, e)
            
            

class Channel:
    """Channel to communicate with another process using the specified transport.
    """
    def __init__( self, transport, request_handler ):
        self.transport = transport
        self.request_handler = request_handler
        self._stop_waiting = False
        self._next_request_id = 0
        self._pending_requests_by_id = {}
        self.logger = logging.getLogger( 'JsonRPC' )

    def close( self ):
        self.stop_waiting_for_requests()
        self.transport.close()

    def _remove_processed_requests( self, request_ids ):
        """Removes all pending request from the head of the list."""
        while request_ids:
            if request_ids[0] in self._pending_requests_by_id:
                return request_ids
            del request_ids[0]
        return request_ids

    def wait_for_requests( self, request_ids = None ):
        """Wait for incomming requests and reply to send request.

           If a PARSE_ERROR occurs then the connection is interrupted as it is assummed that the peer as trouble decoding the emitted JSON.

           Parameters:
           request_ids: if specified, exit as soon as reply have been recieved for the specified request. If requests
                        have already been replied to then returns immediately.

           Exceptions: JsonRPCError on incoming request that can not be parsed/decoded.
        """
        if request_ids:
            request_ids = self._remove_processed_requests( list(request_ids) )
            exit_on_pending_requests_completion = True
        else:
            exit_on_pending_requests_completion = False
        
        while not self._stop_waiting:
            if exit_on_pending_requests_completion:
                request_ids = self._remove_processed_requests( request_ids )
                if not request_ids: # all pending request have been processed
                    break
                
            encoded_json = self.transport.read_line()
            try:
                request = json.loads( encoded_json )
            except ValueError, e:
                self.logger.error( 'Failed to decoded incomming message', exc_info = True )
                self.reply_with_error_and_raise( None, e, PARSE_ERROR )
            else:
                if not isinstance(request, dict):
                    self.logger.error( 'Invalid incoming request: not a JSON object: %s' % request )
                    self.reply_with_error_and_raise( None, INVALID_REQUEST.detail(request) )
                else:
                    request_id = request.get('id')
                    method = request.get('method')
                    has_result = 'result' in request
                    has_error = 'error' in request
                    is_reply = has_result or has_error
                    is_valid_request_id = request_id is None or isinstance(request_id,int)
                    is_valid_method = isinstance(method, basestring)
                    if is_reply:
                        if not is_valid_request_id or request_id not in self._pending_requests_by_id:
                            self.reply_with_error_and_raise( None, INVALID_REPLY.detail({'id': request_id}) )
                        elif has_result and has_error:
                            self.reply_with_error_and_raise( None, INVALID_REPLY.detail('Reply can not have both a result and an error.') )
                        else:
                            reply_handler = self._pending_requests_by_id[request_id]
                            del self._pending_requests_by_id[request_id]
                            if has_result:
                                reply_handler( request_id, request['result'], None, self )
                            else:
                                reply_handler( request_id, None, request['error'], self )
                    elif not is_valid_method or not is_valid_request_id:
                        self.reply_with_error_and_raise( is_valid_request_id and request_id or None, INVALID_REQUEST )
                    elif request_id is None: # notification
                        self.request_handler.on_notification( method, request.get('params'), self )
                    else: # request
                        self.request_handler.onRequest( request_id, method, request.get('params'), self )

    def stop_waiting_for_requests( self ):
        self._stop_waiting = True

    def execute_synchronous_request( self, method, parameter ):
        """Execute the specified request and returns its result.

           Notes: Incoming notification with continue to be dispatched to
                  registered handlers while waiting for the reply.

           Parameters:
           method: Name of the method to invoke
           parameter: dict to pass as parameter (JSON object)

           Returns: request result object

           Exceptions:
           - JsonRPCError raised if the request fails (I/O Error or error reply).
           - OverflowError, TypeError, ValueError on JSON encoding error when sending the message,
             meaning that an invalid 'parameter' was passed.
        """
        class SynchronousReplyHandler:
            def __init__( self ):
                self.reply = None
                
            def __call__( self, request_id, result, error, channel ) :
                self.reply = (result, error)

        handler = SynchronousReplyHandler()
        request_id = self.send_request( method, parameter, handler )
        try:
            self.wait_for_requests( request_ids = [request_id] )
        except IOError, e:
            handler( request_id, None, CONNECTION_LOST.detail_from_exception( e ), self )
        except EOFError, e:
            handler( request_id, None, CONNECTION_LOST.detail_from_exception( e ), self )
        # @todo edge case not handled: if exited waiting because stop_waiting_for_requests() was called...
        assert handler.reply is not None
        if handler.reply[1]: # got an error
            raise handler.reply[1]
        return handler.reply[0]

    def send_request( self, method, parameter, reply_handler ):
        """Sends the specified request to the peer.

           Parameters:
           method: Name of the method to invoke
           parameter: dict to pass as parameter (JSON object)
           reply_handler: call back, must be a callable (request_id, result, error, channel)
                          Notes: the reply_handler may be called before returning from send_request() if
                          an I/O error occurs while sending the request.

           Returns: the request id
        """
        self._next_request_id += 1
        request_id = self._next_request_id
        request = {
            'method': method,
            'id': request_id
            }
        if parameter is not None:
            request['params'] = parameter
        try:
            self._send_message( None, request )
        except IOError, e:
            reply_handler( request_id, None, CONNECTION_LOST.detail_from_exception( e ), self )
        except EOFError, e:
            reply_handler( request_id, None, CONNECTION_LOST.detail_from_exception( e ), self )
        else:
            self._pending_requests_by_id[request_id] = reply_handler
        return request_id

    def send_notification( self, method, parameter ):
        request = {
            'method': method,
            }
        if parameter is not None:
            request['params'] = parameter
        self._send_message( None, request )

    def reply( self, request_id, result ):
        reply = {
            'id': request_id,
            'result': result
            }
        self._send_message( request_id, reply )

    def reply_with_error( self, request_id, error ):
        self.logger.debug( 'Replying with error %s:\n%s', error, traceback.format_tb() )
        reply = {
            'id': request_id,
            'error': error.data
            }
        self._send_message( request_id, reply )

    def reply_with_error_and_raise( self, request_id, error ):
        """Reply with the specified error and raise it.
           Used for communication error that put the state of the peers in
           an undetermined state (parse error, invalid request...).
        """
        try:
            self.reply_with_error( request_id, error )
        except Exception, e:
            self.logger.info( 'Failed to reply with error', exc_info = True )
        raise error

    def reply_with_exception( self, request_id, exception, error = None ):
        if error is None:
            error = INTERNAL_ERROR
        self.logger.debug( 'Replying with exception %s', exception, exc_info = True )
        error = error.detail_from_exception( exception )
        reply = {
            'id': request_id,
            'error': error.data
            }
        self._send_message( request_id, reply )

    def _send_message( self, request_id, json_message ):
        json_message['jsonrpc'] = '2.0'
        encoded_json = json.dumps( json_message )
        self.transport.write_line( encoded_json )
