from jsonrpc import SubprocessPipeTransport, Channel
import jsonrpc
import sys
import json
import logging
from cStringIO import StringIO
import time


class TestEntry:
    def __init__( self, name, qualified_name, data = None ):
        self.name = name
        self.qualified_name = qualified_name
        self.data = data and data.copy() or {}
        self.parent = None

class TestSuite(TestEntry):
    def __init__( self, *args, **kwargs ):
        TestEntry.__init__( self, *args, **kwargs )
        self.nested_suites_by_name = {}
        self.test_cases_by_name = {}

    def add_nested_suite( self, suite ):
        assert suite.parent == None
        self.nested_suites_by_name[suite.name] = suite
        suite.parent = self

    def add_test_case( self, test ):
        assert test.parent == None
        self.test_cases_by_name[test.name] = test
        test.parent = self

    def __repr__( self ):
        return 'TestSuite(name="%s", qualified_name="%s")' % (self.name, self.qualified_name)

class TestCase(TestEntry):

    def __repr__( self ):
        return 'TestCase(name="%s", qualified_name="%s")' % (self.name, self.qualified_name)

class TestEvent:
    def __init__( self ):
        pass

class SourceLocation:
    def __init__( self, source, line ):
        self.source = source
        self.line = line

    def __str__( self ):
        return '%s(%d) : ' % (self.source, self.line)

class TestLogEvent(TestEvent):
    def __init__( self, value ):
        TestEvent.__init__( self )
        self.value = value

    def __str__( self ):
        return json.dumps(self.value, indent=4, sort_keys=True)

class TestAssertionEvent(TestEvent):
    def __init__( self, kind, source_location, is_ignored_failure, detail ):
        TestEvent.__init__( self )
        self.kind = kind # fault, assertion
        self.location = source_location
        self.is_ignored_failure = is_ignored_failure
        self.detail = detail
        self.messages = detail.get('message')
        self.predicate = detail.get('name')
        self.detail_data = self.detail.get('data', [])

    def __str__( self ):
        out = StringIO()
        if self.location:
            out.write( str(self.location) )
        else:
            out.write( 'unknown failure location : ' )
        if self.is_ignored_failure:
            out.write( '[failure type: *ignored* %s]\n' % self.kind )
        else:
            out.write( '[failure type: %s]\n' % self.kind )
        if self.messages:
            if len(self.messages) == 1:
                out.write( 'Message: %s\n' % self.messages[0] )
            else:
                out.write( 'Messages:\n' )
                out.write( '\n'.join( self.messages ) )
                out.write( '\n')
        data = self.detail_data
        max_name_length = max( [0] + [len(entry['name']) for entry in data] )
        if self.predicate:
            out.write( 'predicate: %s\n' % self.predicate )
            max_name_length = max( (max_name_length, len('predicate')) )
        for entry in data:
            name, value = entry['name'], entry['value']
            value_as_text = json.dumps( value, indent=4, sort_keys=True)
            out.write( '%-*s: %s\n' % (max_name_length, name, value_as_text) )
        # todo: print composition
        return out.getvalue()


ASSERTION_COUNT_STAT = 'assertions'
FAILED_ASSERTION_COUNT_STAT = 'failedAssertions'
IGNORED_FAILURE_STAT = 'ignoredFailure'

STATUS_PASSED = 'passed'
STATUS_FAILED = 'failed'
STATUS_SKIPPED = 'skipped'

class TestResult:
    def __init__( self, test, status, status_detail = None, events = None ):
        self.test = test
        self.status = status
        self.status_detail = status_detail or {}
        self.events = list(events or ())
        self.statistics = self.status_detail.get('statistics', {})

class TestEventHandler:
    """Strategy used to handle all test events during the test run.
    """
    def __init__( self ):
        pass
    
    def test_started( self, test ):
        """Called when a test is started.
           Parameters:
           test: the TestCase instance of the test being started.
           """
        pass

    def test_ended( self, test_result ):
        """Called when a test ended its execution.
           Parameters:
           test_result: a TestResult instance.
           """
        pass

    # @todo change interface to handle recovery restart...
    def run_started( self, tests ):
        """Called before starting the execution of any tests.
           Parameters:
           tests: sequence of TestCase instance that will be run.
        """
        pass

    def run_ended( self ):
        """Called after the execution of all tests.
        """
        pass


class TestSlaveFactory:
    def __init__( self, path, *args ):
        self.path = path
        self.args = list( args )
        self.args.append( '--test-slave' )
        self._root_suite = None
        self._suites = []
        self._tests = []

    def spawn_slave( self ):
        """Creates a new test slave.
           Returns: new TestSlave instance.
        """
        return TestSlave( self, self.path, *self.args )

    def release_slave( self, test_slave ):
        """Terminates the slave process."""
        test_slave.channel.close()

    def make_suite( self, parent_suite, name, qualified_name, data ):
        if parent_suite is None:
            if self._root_suite is None:
                self._root_suite = TestSuite( name, qualified_name, data )
            return self._root_suite
        suite = parent_suite.nested_suites_by_name.get( name )
        if suite is None:
            suite = TestSuite( name, qualified_name, data )
            parent_suite.add_nested_suite( suite )
        return suite

    def make_test_case( self, parent_suite, name, qualified_name, data ):
        test = parent_suite.test_cases_by_name.get( name )
        if test is None:
            test = TestCase( name, qualified_name, data )
            parent_suite.add_test_case( test )
        return test

class TestSlave:
    def __init__( self, factory, path, *args ):
        self.factory = factory
        args = list(args)
        args.insert(0, path )
        transport = jsonrpc.SubprocessPipeTransport( args )
        handler = jsonrpc.RequestHandler()
        handler.register_notification_method( 'testStarted', self._onTestStarted )
        handler.register_notification_method( 'testEnded', self._onTestEnded )
        self.channel = jsonrpc.Channel( transport, handler )
        self.suites_by_id = {}
        self.tests_by_id = {}
        self.ids_by_test = {}
        self.event_handler = None
        self.running_tests = {}     # {test: active_count}
        self.ended_tests = []

    def release( self ):
        self.factory.release_slave( self )

    def get_all_tests( self ):
        """Get the list of all test cases declared in the test slave.
           Returns: list(TestCase)
        """
        self.event_handler = None
        result = self.channel.execute_synchronous_request( 'listTests', {} )
        
        suites_by_id = {}
        tests_by_id = {}
        for entry in result:
            type = entry['type']
            entry_id = entry['id']
            qualified_name = entry.get('qualifiedName')
            name = entry['name']
            data = entry.get('data')
            parent = entry.get('parent')
            if parent is not None:
                parent_suite = suites_by_id[parent]
                if qualified_name is None:
                    qualified_name = parent_suite.qualified_name
                    if not qualified_name.endswith('/'):
                        qualified_name += '/'
                    qualified_name += name
            else:
                parent_suite = None
                if qualified_name is None:
                    qualified_name = '/'
            if type == 'test':
                if parent_suite is None:
                    raise ValueError( 'Test case "%s" must have a parent suite' % qualified_name )
                test = self.factory.make_test_case( parent_suite, name, qualified_name, data )
                tests_by_id[entry_id] = test
            elif type == 'suite':
                suite = self.factory.make_suite( parent_suite, name, qualified_name, data )
                suites_by_id[entry_id] = suite
            else:
                raise ValueError( 'Invalid entry type: "%s"' % type )
        self.suites_by_id = suites_by_id
        self.tests_by_id = tests_by_id

        self.ids_by_test = {}
        for id, test in self.tests_by_id.iteritems():
            self.ids_by_test[test] = id

        return self.tests_by_id.values()

    def run_tests( self, tests, test_result_handler ):
        """Run the list of specified test and send result notification to the specified handler.
           Parameters:
           tests: list(TestCase)
           test_result_handler: a sub-class of TestEventHandler.
                                Notes: Only testStarted() and testEnded() notification will be send.

           Returns: tuple(ended_tests, running_tests, remaining_tests), where
           - ended_tests is a list(TestCase) of test cases that have been executed,
           - running_tests is a list(TestCase) of test cases that where started but not ended when the slave terminated
             (e.g. it is the list of the that were being executed when the slave crashed),
           - remaining_tests is a  list(TestCase) of test cases that were requested but not ended (e.g. includes running_tests).
        """
        if not self.tests_by_id and tests:
            try:
                self.get_all_tests() # initialize tests/id dict
            except jsonrpc.JsonRPCError, e:
                return ([], [], tests)
        tests_ids = [ self.ids_by_test[test] for test in tests ]
        self.running_tests = {}     # {test: active_count}
        self.ended_tests = []
        remaining_tests = {}   # {test: active_count}
        for test in tests:
            remaining_tests[test] = remaining_tests.get(test, 0) + 1
        
        self.event_handler = test_result_handler
        try:
            result = self.channel.execute_synchronous_request( 'runTests', {'tests': tests_ids} )
        except jsonrpc.JsonRPCError, e:
            pass
        self.event_handler = None

        running_tests = []
        for test, count in self.running_tests.iteritems():
            running_tests.extend( [test] * count )
        for test in self.ended_tests:
            if test in remaining_tests:
                count = remaining_tests[test] - 1
                if count > 0:
                    remaining_tests[test] = count
                else:
                    del remaining_tests[test]
        remaining_tests_list = []
        for test, count in remaining_tests.iteritems():
            remaining_tests_list.extend( [test] * count )
        return (self.ended_tests, running_tests, remaining_tests_list)

    def _handleRunTestsReply( self, request_id, result, error, channel ):
        if error:
            print 'Failed to get list of test:'
            print error
            channel.close()
        else:
            self.event_handler.run_ended()
            time.sleep( 0.1 )
            channel.close()

    def _onTestStarted( self, parameter, channel ):
        test_id = parameter['id']
        data = parameter.get('data', {})
        test = self.tests_by_id[test_id]

        self.running_tests[test] = self.running_tests.get(test, 0) + 1
        
        self.event_handler.test_started( test )

    def _onTestEnded( self, parameter, channel ):
        test_id = parameter['id']
        test = self.tests_by_id[test_id]

        if test in self.running_tests: # should we emit a warning if test was not started?
            count = self.running_tests[test] - 1
            if count > 0:
                self.running_tests[test] = count
            else:
                del self.running_tests[test]

        # @todo handle invalid parameter error so test is not rerun again...        
        events = parameter.get('events', [] )
        test_status = parameter['status']
        status = test_status['status']
        events = []
        for event in parameter.get('events', [] ):
            if event['type'] == 'log':
                translated_event = TestLogEvent( event['data'] )
            elif event['type'] == 'assertion':
                kind = event['kind']
                location = event['location']
                is_ignored_failure = event.get('ignored_failure', False)
                detail = event.get('detail', {})
                source_location = SourceLocation( location['source'], location['line'] )
                translated_event = TestAssertionEvent( kind, source_location, is_ignored_failure, detail )
            else:
                raise ValueError( 'Unknown event type: "%s"' % event['type'] )
            events.append( translated_event )

        self.ended_tests.append( test )
            
        test_result = TestResult( test, status, status_detail = test_status, events = events )
        self.event_handler.test_ended( test_result )
       

class CompilerTestEventHandler(TestEventHandler):
    def __init__( self ):
        TestEventHandler.__init__( self )
        self.test_results_with_events = []
        self._reset_run_statistics()

    def _reset_run_statistics( self ):
        self.test_ended_count = 0
        self.test_failed_count = 0
        self.ignored_failure_count = 0
        self.total_assertion_count = 0

    def get_exit_code( self ):
        return self.test_failed_count > 0 and 1 or 0
    
    def test_started( self, test ):
        print 'Testing %s:' % test.qualified_name,

    def test_ended( self, test_result ):
        status_text = {'passed': 'OK',
                       'failed': 'FAIL',
                       'skipped': 'SKIP'}
        
        assertions = test_result.statistics.get( ASSERTION_COUNT_STAT, 0 )
        failures = test_result.statistics.get( FAILED_ASSERTION_COUNT_STAT, 0 )
        ignored_failures = test_result.statistics.get( IGNORED_FAILURE_STAT, 0 )
        out = StringIO()
        out.write( status_text[test_result.status] )
        out.write( ' (' )
        if failures:
            out.write( '%d assertion(s) failed' % failures )
            if assertions:
                out.write( '/' )
        if assertions:
            out.write( '%d %s' % (assertions, assertions > 1 and 'assertions' or 'assertion' ) )
        if ignored_failures:
            out.write( ', %d ignored failures' % ignored_failures )
        out.write( ')')
        print out.getvalue()
        
        if test_result.events and test_result.status == 'failed':
            self.test_results_with_events.append( test_result )

        self.test_ended_count += 1
        if test_result.status == STATUS_FAILED:
            self.test_failed_count += 1
        for event in test_result.events:
            if event.is_ignored_failure:
                self.ignored_failure_count += 1
        self.total_assertion_count += assertions

    # @todo change interface to handle recovery restart...
    def run_started( self, tests ):
        """Called before starting the execution of any tests.
           Parameters:
           tests: sequence of TestCase instance that will be run.
        """
        self._reset_run_statistics()

    def run_ended( self ):
        """Called after the execution of all tests.
           Parameters:
           tests: sequence of TestCase instance that should have been run.
        """
        # Statistics summary
        out = StringIO()
        if self.test_failed_count == 0:
            out.write( 'All %d test(s) passed' % self.test_ended_count )
        else:
            out.write( 'Failure report:\n' )
            # Detail of previous failures
            # @todo some tests may be missing if they failed without generating events (is that possible?)
            for test_result in self.test_results_with_events:
                out.write( '-> %s:\n' % test_result.test.qualified_name )
                for event in test_result.events:
                    if isinstance( event, TestLogEvent ):
                        out.write( 'Log: %s\n' % str(event) )
                    elif isinstance( event, TestAssertionEvent ):
                        out.write( str(event) )
                    else:
                        raise ValueError( 'Unknown event type: "%s"' % repr(event) )
                out.write( '\n' )
            out.write( '%d/%d tests passed, %d tests failed' % (
                self.test_ended_count - self.test_failed_count,
                self.test_ended_count,
                self.test_failed_count ) )
        if self.ignored_failure_count > 0:
            out.write( ', %d ignored failures' % self.ignored_failure_count )
        out.write( ' (%d %s).\n' % ( self.total_assertion_count, self.total_assertion_count > 1 and 'assertions' or 'assertion' ) )
        print out.getvalue()


class TestRunner:
    def run( self, slave_factory, test_result_handler ):
        test_slave = slave_factory.spawn_slave()
        remaining_tests = test_slave.get_all_tests()
        test_result_handler.run_started( remaining_tests )
##        print 'to_run:', remaining_tests
        while remaining_tests:
            if test_slave is None:
                test_slave = slave_factory.spawn_slave()
                
            ended_tests, running_tests, remaining_tests = test_slave.run_tests( remaining_tests, test_result_handler )
            # Generate failure events for test that were running while the crash occurred
            for test in running_tests:
                event_detail = {
                    'message': ['Process crashed while running the test']
                    }
                event = TestAssertionEvent( 'fault', None, False, event_detail )
                test_result = TestResult( test, 'failed', events = [event] )
                test_result_handler.test_ended( test_result )
                remaining_tests.remove( test )
##            print 'ended:', ended_tests
##            print 'running:', running_tests
##            print 'remaining_tests:', remaining_tests
                
            test_slave.release()
            test_slave = None
        test_result_handler.run_ended()
        

if __name__ == '__main__':
    import sys
    import os.path
    path = sys.argv[1]

    log_path = 'log.txt'
    if os.path.exists( log_path ):
        os.unlink( log_path )
    logging.basicConfig(filename=log_path,level=logging.DEBUG,)

    result_handler = CompilerTestEventHandler()
    slave_factory = TestSlaveFactory( path )
    successful = TestRunner().run( slave_factory, result_handler )
    sys.exit( not successful and 1 or result_handler.get_exit_code() )

    
